/*
 * TracingClassLoader.java
 *
 * Created on April 2, 2002, 6:53 PM
 */

package name.dsouflis.aop.tracing;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.Repository;
import org.apache.bcel.Constants;
import org.apache.bcel.generic.*;

/** Replacement for the system ClassLoader that instruments some classes on load-time with output
 * statements that trace entering and exiting.
 * @author dsouflis
 * @version $Revision: 1.3 $
 */
public class TracingClassLoader extends ClassLoader {

    /** The filter determining the classes to instrument */
    private ClassMethodFilter filter;

    /** Creates new TracingClassLoader
     * @param filter The filter that determines class names of classes to instrument.
     */
    public TracingClassLoader(ClassMethodFilter filter) {
        this.filter=filter;
    }

    /** Creates new TracingClassLoader that traces everything by default
     */
    public TracingClassLoader() {
        this.filter=new ClassMethodFilter() {
            public boolean classHasAdvice(String className) {
                return true;
            }
            public boolean methodHasAdvice(String className, String methodName, String signature) {
                return true;
            }
        };
    }

    /** Loads the class, adding instrumentation if necessary.<br>
     * Delegates to the parent ClassLoader system classes.
     * @param name Guess what.
     * @param resolve Whether the class should be resolved.
     * @throws ClassNotFoundException If the class cannot be found.
     * @return The loaded class.
     */
    protected Class loadClass(String name, boolean resolve) throws java.lang.ClassNotFoundException {
        Class theClass=findLoadedClass(name);
        if(theClass!=null) {
            return theClass;
        }
        if(name.startsWith("java")) {
            return getParent().loadClass(name);
        }
        theClass=findClass(name);
        if(theClass==null) {
            return null;
        }
        if(resolve) {
            resolveClass(theClass);
        }
        return theClass;
    }

    /** Defines the class.
     * @param className Guess what.
     * @return The loaded class.
     */
    protected Class findClass(String className) {
        JavaClass clazz=classWithTracing(filter,className);
        if(clazz==null) {
            return null;
        }
        byte[] b = clazz.getBytes();
        return defineClass(className, b, 0, b.length);
    }

    /** Returns a JavaClass, instrumented if its name is contained in the list.
     * @param filter The filter that determines class names of classes to instrument.
     * @param className The class to modify.
     * @return The resulting JavaClass.
     */
    public static JavaClass classWithTracing(ClassMethodFilter filter, String className) {
        JavaClass clazz;
        try {
            clazz= Repository.lookupClass(className);
        } catch (ClassNotFoundException e) {
            clazz=null;
        }
        if(clazz==null) {
            return null;
        }
        if(filter.classHasAdvice(className)) {
            clazz=addTracing(filter, clazz);
        }
        return clazz;
    }

    /** Adds tracing instrumentation to a JavaClass.
     * @param filter The filter that determines class names of classes to instrument.
     * @param clazz The JavaClass.
     * @return The resulting JavaClass.
     */
    public static JavaClass addTracing(ClassMethodFilter filter, JavaClass clazz) {
        ClassGen cg=new ClassGen(clazz);
        ConstantPoolGen cp = cg.getConstantPool();
        Method[] methods=cg.getMethods();
        int methodNum=methods.length;
        for(int i=0; i<methodNum; i++) {
            addMethodWithTracing(filter, cg, i);
        }
        clazz=cg.getJavaClass();
        return clazz;
    }

    /** Adds a method that wraps a method of the original class, unless the
     * method is a constructor, or is abstract.
     * @param filter The filter that determines class names of classes to instrument.
     * @param cg The ClassGen that represents the class under construction.
     * @param i The method of the original class to wrap.
     */
    private static void addMethodWithTracing(ClassMethodFilter filter, ClassGen cg, int i) {
        ConstantPoolGen cp = cg.getConstantPool();
        int out     = cp.addFieldref("java.lang.System", "out",
        "Ljava/io/PrintStream;");
        int printlnString = cp.addMethodref("java.io.PrintStream", "println",
        "(Ljava/lang/String;)V");
        int printString = cp.addMethodref("java.io.PrintStream", "print",
        "(Ljava/lang/String;)V");
        int currentThread = cp.addMethodref("java.lang.Thread", "currentThread",
        "()Ljava/lang/Thread;");

        Method m=cg.getMethodAt(i);
        Code   code  = m.getCode();
        int    flags = m.getAccessFlags();
        String name  = m.getName();
        String signature=m.getSignature();

        if(name.equals("<init>") || code==null) {
            return;
        }
        if(!filter.methodHasAdvice(cg.getClassName(),name,signature)) {
            return;
        }

        String hiddenName="hidden$"+name;
        int nameIndex=cp.addUtf8(hiddenName);
        m.setNameIndex(nameIndex);

        if(m.isPublic()) {
            int newFlags=(flags&~Constants.ACC_PUBLIC)| Constants.ACC_PRIVATE;
            m.setAccessFlags(newFlags);
        }

        cg.setMethodAt(m,i);

        int originalMethodIndex=cp.addMethodref(cg.getClassName(),name,signature);
        int hiddenMethodIndex=cp.addMethodref(cg.getClassName(),hiddenName,signature);
        InstructionList il=new InstructionList();

        Type[] argTypes=Type.getArgumentTypes(signature);
        Type returnType=Type.getReturnType(signature);
        int argNum=argTypes.length;
        LocalVariableTable lvt=m.getLocalVariableTable();
        int argOffset=0;
        MethodGen newMethod=new MethodGen(
        flags,
        returnType,
        argTypes,
        null,
        name,
        cg.getClassName(),
        il,
        cp);
        InstructionHandle startHd=null;
        InstructionHandle endHd=null;
        InstructionHandle exceptHd=null;

        //System.out.print(java.lang.Thread.currentThread());
        startHd=il.append(new GETSTATIC(out));
        il.append(new INVOKESTATIC(currentThread));
        il.append(printInstruction(cp, Type.OBJECT,"print"));

        if(!m.isStatic()) {
            argOffset++;
            //System.out.print(this);
            il.append(new GETSTATIC(out));
            il.append(new ALOAD(0));
            il.append(printInstruction(cp,Type.OBJECT,"print"));
            //System.out.print(".");
            il.append(new GETSTATIC(out));
            il.append(new PUSH(cp, "."));
            il.append(new INVOKEVIRTUAL(printString));
        }
        //System.out.print("<<name>>");
        il.append(new GETSTATIC(out));
        il.append(new PUSH(cp, name));
        il.append(new INVOKEVIRTUAL(printString));
        //System.out.print("(");
        il.append(new GETSTATIC(out));
        il.append(new PUSH(cp, "("));
        il.append(new INVOKEVIRTUAL(printString));
        //push args, one by one, invoking appropriate print
        for(int iarg=0; iarg<argNum; iarg++) {
            if(iarg!=0) {
                //System.out.print(",");
                il.append(new GETSTATIC(out));
                il.append(new PUSH(cp, ","));
                il.append(new INVOKEVIRTUAL(printString));
            }
            //System.out.print(<<arg[iarg]>>);
            il.append(new GETSTATIC(out));
            Instruction instr=InstructionFactory.createLoad(argTypes[iarg],slotOfArg(iarg,argTypes)+argOffset);
            il.append(instr);
            il.append(printInstruction(cp,argTypes[iarg],"print"));
        }
        //System.out.println(")");
        il.append(new GETSTATIC(out));
        il.append(new PUSH(cp, ")"));
        il.append(new INVOKEVIRTUAL(printlnString));

        //T temp=<<originalMethod>>(<<original arguments>>);
        if(!m.isStatic()) {
            il.append(new ALOAD(0));
            argOffset=1;
        }
        for(int iarg=0; iarg<argNum; iarg++) {
            Instruction instr=InstructionFactory.createLoad(argTypes[iarg],slotOfArg(iarg,argTypes)+argOffset);
            il.append(instr);
        }
        if(m.isStatic()){
            il.append(new INVOKESTATIC(hiddenMethodIndex));
        } else {
            il.append(new INVOKESPECIAL(hiddenMethodIndex));
        }

        //System.out.print(java.lang.Thread.currentThread());
        il.append(new GETSTATIC(out));
        il.append(new INVOKESTATIC(currentThread));
        il.append(printInstruction(cp,Type.OBJECT,"print"));
        //System.out.print("<<name>>");
        il.append(new GETSTATIC(out));
        il.append(new PUSH(cp, name));
        il.append(new INVOKEVIRTUAL(printString));

        if(returnType!=Type.VOID) {
            il.append(InstructionFactory.createStore(returnType,slotOfArg(argNum,argTypes)+argOffset));
            //System.out.print("=>");
            il.append(new GETSTATIC(out));
            il.append(new PUSH(cp, "=>"));
            il.append(new INVOKEVIRTUAL(printString));
            //System.out.println(temp);
            il.append(new GETSTATIC(out));
            il.append(InstructionFactory.createLoad(returnType,slotOfArg(argNum,argTypes)+argOffset));
            il.append(printInstruction(cp,returnType,"println"));
            //return temp;
            il.append(InstructionFactory.createLoad(returnType,slotOfArg(argNum,argTypes)+argOffset));
            endHd=il.append(InstructionFactory.createReturn(Type.getReturnType(signature)));
        } else {
            //System.out.print("=>(void)");
            il.append(new GETSTATIC(out));
            il.append(new PUSH(cp, "=>(void)"));
            il.append(new INVOKEVIRTUAL(printlnString));
            endHd=il.append(new RETURN());
        }

        exceptHd=il.append(InstructionFactory.createStore(Type.OBJECT,slotOfArg(argNum,argTypes)+argOffset));

        //System.out.print(java.lang.Thread.currentThread());
        il.append(new GETSTATIC(out));
        il.append(new INVOKESTATIC(currentThread));
        il.append(printInstruction(cp,Type.OBJECT,"print"));
        //System.out.print("<<name>>");
        il.append(new GETSTATIC(out));
        il.append(new PUSH(cp, name));
        il.append(new INVOKEVIRTUAL(printString));

        //System.out.print(" =>threw exception ");
        il.append(new GETSTATIC(out));
        il.append(new PUSH(cp, "=>threw exception "));
        il.append(new INVOKEVIRTUAL(printString));

        il.append(new GETSTATIC(out));
        il.append(InstructionFactory.createLoad(Type.OBJECT,slotOfArg(argNum,argTypes)+argOffset));
        il.append(printInstruction(cp,Type.OBJECT,"println"));

        il.append(InstructionFactory.createLoad(Type.OBJECT,slotOfArg(argNum,argTypes)+argOffset));
        il.append(new ATHROW());

        newMethod.setInstructionList(il);
        CodeExceptionGen exc=newMethod.addExceptionHandler(startHd,endHd,exceptHd,(ObjectType)null);
        newMethod.setMaxLocals();
        newMethod.setMaxStack(Math.max(4,code.getMaxStack()));
        cg.addMethod(newMethod.getMethod());
        il.dispose();
    }

    /** Computes the local slot of an argument, not taking into account
     * the presence of <CODE>this</CODE>.
     * @param n The argument ordinal number.
     * @param types The types of the arguments.
     * @return The argument slot, not taking into account <CODE>this</CODE>.
     */
    private static int slotOfArg(int n, Type[] types) {
        int s=0;
        for(int i=0; i<n; i++) {
            if(types[i]==Type.LONG || types[i]==Type.DOUBLE) {
                s+=2;
            } else {
                s++;
            }
        }
        return s;
    }

    /** Creates a print instruction of the appropriate type.
     * @param cp The ConstantPoolGen that represents the constant pool of the class
     * under construction.
     * @param t The type of the value to print.
     * @param method The method to create calling instruction for (<CODE>print</CODE>
     * or <CODE>println</CODE>.
     * @return The instruction.
     */
    private static Instruction printInstruction(ConstantPoolGen cp, Type t, String method) {
        //method is either "print" or "println"
        int printX;
        if(t==Type.BOOLEAN||t==Type.INT||t==Type.SHORT||t==Type.CHAR||t==Type.BYTE
        ||t==Type.LONG||t==Type.FLOAT||t==Type.DOUBLE) {
            printX=cp.addMethodref(
            "java.io.PrintStream",
            method,
        Type.getMethodSignature(Type.VOID,new Type[]{t}));
        } else { //Type.OBJECT and all others
            printX=cp.addMethodref(
            "java.io.PrintStream",
            method,
        Type.getMethodSignature(Type.VOID,new Type[]{Type.OBJECT}));
        }
        return new INVOKEVIRTUAL(printX);
    }
}